# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This package contains all battery-related application classes, including
representations of InsertionElectrodes and ConversionElectrodes.
"""
