# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

"""
This package is an overarching package for subpackages for specific materials
applications.
"""
